@@ -1,88 +1,99 @@
 from __future__ import absolute_import, division, print_function
 
 import logging
-from itertools import groupby
-from collections import namedtuple, OrderedDict
-from math import ceil
+import os
+import sys
+import uuid
 import warnings
+from collections import namedtuple, OrderedDict
+from itertools import groupby, repeat
+from math import ceil
 
-import pandas
 import numpy
+import pandas
 import xarray
+
+try:
+    import SharedArray as sa
+except ImportError:
+    pass
 from affine import Affine
 from dask import array as da
 
-from ..config import LocalConfig
-from ..compat import string_types
-from ..index import index_connect
-from ..storage.storage import RasterDatasetSource, reproject_and_fuse
-from ..utils import geometry, intersects, data_resolution_and_offset
+try:
+    from pathos.threading import ThreadPool
+except ImportError:
+    pass
+from six.moves import zip
+
+from datacube.config import LocalConfig
+from datacube.compat import string_types
+from datacube.storage.storage import reproject_and_fuse
+from datacube.utils import geometry, intersects, data_resolution_and_offset
 from .query import Query, query_group_by, query_geopolygon
+from ..index import index_connect
+from ..drivers import new_datasource
 
 _LOG = logging.getLogger(__name__)
-
+THREADING_REQS_AVAILABLE = ('SharedArray' in sys.modules and 'pathos.threading' in sys.modules)
 
 Group = namedtuple('Group', ['key', 'datasets'])
 
 
-def _xarray_affine(obj):
-    dims = obj.crs.dimensions
-    xres, xoff = data_resolution_and_offset(obj[dims[1]].values)
-    yres, yoff = data_resolution_and_offset(obj[dims[0]].values)
-    return Affine.translation(xoff, yoff) * Affine.scale(xres, yres)
-
-
-def _xarray_extent(obj):
-    return obj.geobox.extent
-
-
-def _xarray_geobox(obj):
-    dims = obj.crs.dimensions
-    return geometry.GeoBox(obj[dims[1]].size, obj[dims[0]].size, obj.affine, obj.crs)
-
-
-xarray.Dataset.geobox = property(_xarray_geobox)
-xarray.Dataset.affine = property(_xarray_affine)
-xarray.Dataset.extent = property(_xarray_extent)
-xarray.DataArray.geobox = property(_xarray_geobox)
-xarray.DataArray.affine = property(_xarray_affine)
-xarray.DataArray.extent = property(_xarray_extent)
-
-
 class Datacube(object):
     """
     Interface to search, read and write a datacube.
 
-    :type index: datacube.index._api.Index
+    :type index: datacube.index.index.Index
     """
-    def __init__(self, index=None, config=None, app=None):
+
+    def __init__(self,
+                 index=None,
+                 config=None,
+                 app=None,
+                 env=None,
+                 validate_connection=True):
         """
         Create the interface for the query and storage access.
 
         If no index or config is given, the default configuration is used for database connection.
 
         :param Index index: The database index to use.
+        :type index: :py:class:`datacube.index.Index` or None.
 
-            Can be created by :class:`datacube.index.index_connect`.
-
-        :param LocalConfig config: A config object or a path to a config file that defines the connection.
+        :param Union[LocalConfig|str] config: A config object or a path to a config file that defines the connection.
 
             If an index is supplied, config is ignored.
         :param str app: A short, alphanumeric name to identify this application.
 
             The application name is used to track down problems with database queries, so it is strongly
             advised that be used.  Required if an index is not supplied, otherwise ignored.
+
+        :param str env: Name of the datacube environment to use.
+            ie. the section name in any config files. Defaults to 'datacube' for backwards
+            compatibility with old config files.
+
+            Allows you to have multiple datacube instances in one configuration, specified on load,
+            eg. 'dev', 'test' or 'landsat', 'modis' etc.
+
+        :param bool validate_connection: Should we check that the database connection is available and valid
+
         :return: Datacube object
+
         """
+        def normalise_config(config):
+            if config is None:
+                return LocalConfig.find(env=env)
+            if isinstance(config, string_types):
+                return LocalConfig.find([config], env=env)
+            return config
+
         if index is None:
-            if config is not None:
-                if isinstance(config, string_types):
-                    config = LocalConfig.find([config])
-                self.index = index_connect(config, application_name=app)
-            else:
-                self.index = index_connect(application_name=app)
-        else:
-            self.index = index
+            index = index_connect(normalise_config(config),
+                                  application_name=app,
+                                  validate_connection=validate_connection)
+
+        self.index = index
 
     def list_products(self, show_archived=False, with_pandas=True):
         """
@@ -92,7 +103,7 @@ class Datacube(object):
         :param with_pandas: return the list as a Pandas DataFrame, otherwise as a list of dict.
         :rtype: pandas.DataFrame or list(dict)
         """
-        rows = [datatset_type_to_row(dataset_type) for dataset_type in self.index.products.get_all()]
+        rows = [dataset_type_to_row(dataset_type) for dataset_type in self.index.products.get_all()]
         if not with_pandas:
             return rows
 
@@ -133,8 +144,8 @@ class Datacube(object):
         return measurements
 
     #: pylint: disable=too-many-arguments, too-many-locals
-    def load(self, product=None, measurements=None, output_crs=None, resolution=None, resampling=None, stack=False,
-             dask_chunks=None, like=None, fuse_func=None, align=None, datasets=None, **query):
+    def load(self, product=None, measurements=None, output_crs=None, resolution=None, resampling=None,
+             dask_chunks=None, like=None, fuse_func=None, align=None, datasets=None, use_threads=False, **query):
         """
         Load data as an ``xarray`` object.  Each measurement will be a data variable in the :class:`xarray.Dataset`.
 
@@ -150,8 +161,8 @@ class Datacube(object):
 
             See :meth:`list_products` for the list of products with their names and properties.
 
-            A product can also be selected by searched using fields, but must only match one product.
-            ::
+            A product can also be selected by searching using fields, but must only match one product.
+            For example::
 
                 platform='LANDSAT_5',
                 product_type='ndvi'
@@ -189,23 +200,11 @@ class Datacube(object):
 
             For data that has different values for the scene overlap the requires more complex rules for combining data,
             such as GA's Pixel Quality dataset, a function can be provided to the merging into a single time slice.
-            ::
 
-                def pq_fuser(dest, src):
-                    valid_bit = 8
-                    valid_val = (1 << valid_bit)
+            See :func:`datacube.helpers.ga_pq_fuser` for an example implementation.
 
-                    no_data_dest_mask = ~(dest & valid_val).astype(bool)
-                    np.copyto(dest, src, where=no_data_dest_mask)
-
-                    both_data_mask = (valid_val & dest & src).astype(bool)
-                    np.copyto(dest, src & dest, where=both_data_mask)
 
         **Output**
-            If the `stack` argument is supplied, the returned data is stacked in a single ``DataArray``.
-            A new dimension is created with the name supplied.
-            This requires all of the data to be of the same datatype.
-
             To reproject or resample the data, supply the ``output_crs``, ``resolution``, ``resampling`` and ``align``
             fields.
 
@@ -249,13 +248,6 @@ class Datacube(object):
 
             Default is (0,0)
 
-        :param stack: The name of the new dimension used to stack the measurements.
-            If provided, the data is returned as a :class:`xarray.DataArray` rather than a :class:`xarray.Dataset`.
-
-            If only one measurement is returned, the dimension name is not used and the dimension is dropped.
-
-        :type stack: str or bool
-
         :param dict dask_chunks:
             If the data should be lazily loaded using :class:`dask.array.Array`,
             specify the chunking size in each output dimension.
@@ -281,41 +273,30 @@ class Datacube(object):
             Optional. If this is a non-empty list of :class:`datacube.model.Dataset` objects, these will be loaded
             instead of performing a database lookup.
 
+        :param bool use_threads:
+            Optional. If this is set to True, IO will be multi-thread.
+            May not work for all drivers due to locking/GIL.
+
+            Default is False.
+
         :param int limit:
             Optional. If provided, limit the maximum number of datasets
             returned. Useful for testing and debugging.
 
-        :return: Requested data in a :class:`xarray.Dataset`.
-            As a :class:`xarray.DataArray` if the ``stack`` variable is supplied.
-
-        :rtype: :class:`xarray.Dataset` or :class:`xarray.DataArray`
+        :return: Requested data in a :class:`xarray.Dataset`
+        :rtype: :class:`xarray.Dataset`
         """
+        if 'stack' in query:
+            raise DeprecationWarning("the `stack` keyword argument is not supported anymore, "
+                                     "please apply `xarray.Dataset.to_array()` to the result instead")
+
         observations = datasets or self.find_datasets(product=product, like=like, **query)
         if not observations:
-            return None if stack else xarray.Dataset()
+            return xarray.Dataset()
 
-        if like:
-            assert output_crs is None, "'like' and 'output_crs' are not supported together"
-            assert resolution is None, "'like' and 'resolution' are not supported together"
-            assert align is None, "'like' and 'align' are not supported together"
-            geobox = like.geobox
-        else:
-            if output_crs:
-                if not resolution:
-                    raise RuntimeError("Must specify 'resolution' when specifying 'output_crs'")
-                crs = geometry.CRS(output_crs)
-            else:
-                grid_spec = self.index.products.get_by_name(product).grid_spec
-                if not grid_spec or not grid_spec.crs:
-                    raise RuntimeError("Product has no default CRS. Must specify 'output_crs' and 'resolution'")
-                crs = grid_spec.crs
-                if not resolution:
-                    if not grid_spec.resolution:
-                        raise RuntimeError("Product has no default resolution. Must specify 'resolution'")
-                    resolution = grid_spec.resolution
-                    align = align or grid_spec.alignment
-            geobox = geometry.GeoBox.from_geopolygon(query_geopolygon(**query) or get_bounds(observations, crs),
-                                                     resolution, crs, align)
+        geobox = output_geobox(like=like, output_crs=output_crs, resolution=resolution, align=align,
+                               grid_spec=self.index.products.get_by_name(product).grid_spec,
+                               datasets=observations, **query)
 
         group_by = query_group_by(**query)
         grouped = self.group_datasets(observations, group_by)
@@ -323,31 +304,28 @@ class Datacube(object):
         measurements = self.index.products.get_by_name(product).lookup_measurements(measurements)
         measurements = set_resampling_method(measurements, resampling)
 
-        result = self.load_data(grouped, geobox, measurements.values(),
-                                fuse_func=fuse_func, dask_chunks=dask_chunks)
-        if not stack:
-            return result
-        else:
-            if not isinstance(stack, string_types):
-                stack = 'measurement'
-            return result.to_array(dim=stack)
+        result = self.load_data(grouped, geobox, list(measurements.values()),
+                                fuse_func=fuse_func,
+                                dask_chunks=dask_chunks,
+                                use_threads=use_threads)
+        return result
 
     def product_observations(self, **kwargs):
         warnings.warn("product_observations() has been renamed to find_datasets() and will eventually be removed",
                       DeprecationWarning)
         return self.find_datasets(**kwargs)
 
-    def find_datasets(self, **kwargs):
+    def find_datasets(self, **search_terms):
         """
-        Find datasets for a product.
+        Search the index and return all datasets for a product matching the search terms.
 
-        :param kwargs: see :class:`datacube.api.query.Query`
+        :param search_terms: see :class:`datacube.api.query.Query`
         :return: list of datasets
         :rtype: list[:class:`datacube.model.Dataset`]
 
         .. seealso:: :meth:`group_datasets` :meth:`load_data` :meth:`find_datasets_lazy`
         """
-        return list(self.find_datasets_lazy(**kwargs))
+        return list(self.find_datasets_lazy(**search_terms))
 
     def find_datasets_lazy(self, limit=None, **kwargs):
         """
@@ -367,14 +345,8 @@ class Datacube(object):
         datasets = self.index.datasets.search(limit=limit,
                                               **query.search_terms)
 
-        polygon = query.geopolygon
-        for dataset in datasets:
-            if polygon:
-                # Check against the bounding box of the original scene, can throw away some portions
-                if intersects(polygon.to_crs(dataset.crs), dataset.extent):
-                    yield dataset
-            else:
-                yield dataset
+        for dataset in select_datasets_inside_polygon(datasets, query.geopolygon):
+            yield dataset
 
     @staticmethod
     def product_sources(datasets, group_by):
@@ -410,7 +382,7 @@ class Datacube(object):
         return sources
 
     @staticmethod
-    def create_storage(coords, geobox, measurements, data_func=None):
+    def create_storage(coords, geobox, measurements, data_func=None, use_threads=False):
         """
         Create a :class:`xarray.Dataset` and (optionally) fill it with data.
 
@@ -428,12 +400,20 @@ class Datacube(object):
 
         :param data_func:
             function to fill the storage with data. It is called once for each measurement, with the measurement
-            as an argument. It should return an appropriately shaped numpy array.
+            as an argument. It should return an appropriately shaped numpy array. If not provided, an empty
+            :class:`xarray.Dataset` is returned.
+
+        :param bool use_threads:
+            Optional. If this is set to True, IO will be multi-thread.
+            May not work for all drivers due to locking/GIL.
+
+            Default is False.
 
         :rtype: :class:`xarray.Dataset`
 
         .. seealso:: :meth:`find_datasets` :meth:`group_datasets`
         """
+
         def empty_func(measurement_):
             coord_shape = tuple(coord_.size for coord_ in coords.values())
             return numpy.full(coord_shape + geobox.shape, measurement_['nodata'], dtype=measurement_['dtype'])
@@ -446,8 +426,19 @@ class Datacube(object):
         for name, coord in geobox.coordinates.items():
             result[name] = (name, coord.values, {'units': coord.units})
 
+        def work_measurements(measurement, data_func):
+            return data_func(measurement)
+
+        use_threads = use_threads and THREADING_REQS_AVAILABLE
+
+        if use_threads:
+            pool = ThreadPool(32)
+            results = pool.map(work_measurements, measurements, repeat(data_func))
+        else:
+            results = [data_func(a) for a in measurements]
+
         for measurement in measurements:
-            data = data_func(measurement)
+            data = results.pop(0)
 
             attrs = {
                 'nodata': measurement.get('nodata'),
@@ -471,7 +462,8 @@ class Datacube(object):
         return Datacube.load_data(*args, **kwargs)
 
     @staticmethod
-    def load_data(sources, geobox, measurements, fuse_func=None, dask_chunks=None, skip_broken_datasets=False):
+    def load_data(sources, geobox, measurements, fuse_func=None, dask_chunks=None, skip_broken_datasets=False,
+                  use_threads=False):
         """
         Load data from :meth:`group_datasets` into an :class:`xarray.Dataset`.
 
@@ -494,17 +486,40 @@ class Datacube(object):
             See the documentation on using `xarray with dask <http://xarray.pydata.org/en/stable/dask.html>`_
             for more information.
 
+        :param bool use_threads:
+            Optional. If this is set to True, IO will be multi-thread.
+            May not work for all drivers due to locking/GIL.
+
+            Default is False.
+
         :rtype: xarray.Dataset
 
         .. seealso:: :meth:`find_datasets` :meth:`group_datasets`
         """
+        if use_threads and ('SharedArray' not in sys.modules or 'pathos.threading' not in sys.modules):
+            use_threads = False
+
         if dask_chunks is None:
             def data_func(measurement):
-                data = numpy.full(sources.shape + geobox.shape, measurement['nodata'], dtype=measurement['dtype'])
-                for index, datasets in numpy.ndenumerate(sources.values):
-                    _fuse_measurement(data[index], datasets, geobox, measurement,
-                                      skip_broken_datasets=skip_broken_datasets,
-                                      fuse_func=fuse_func)
+                if not use_threads:
+                    data = numpy.full(sources.shape + geobox.shape, measurement['nodata'], dtype=measurement['dtype'])
+                    for index, datasets in numpy.ndenumerate(sources.values):
+                        _fuse_measurement(data[index], datasets, geobox, measurement, fuse_func=fuse_func,
+                                          skip_broken_datasets=skip_broken_datasets)
+                else:
+                    def work_load_data(array_name, index, datasets):
+                        data = sa.attach(array_name)
+                        _fuse_measurement(data[index], datasets, geobox, measurement, fuse_func=fuse_func,
+                                          skip_broken_datasets=skip_broken_datasets)
+
+                    array_name = '_'.join(['DCCORE', str(uuid.uuid4()), str(os.getpid())])
+                    sa.create(array_name, shape=sources.shape + geobox.shape, dtype=measurement['dtype'])
+                    data = sa.attach(array_name)
+                    data[:] = measurement['nodata']
+
+                    pool = ThreadPool(32)
+                    pool.map(work_load_data, repeat(array_name), *zip(*numpy.ndenumerate(sources.values)))
+                    sa.delete(array_name)
                 return data
         else:
             def data_func(measurement):
@@ -514,25 +529,30 @@ class Datacube(object):
                                         dask_chunks=dask_chunks)
 
         return Datacube.create_storage(OrderedDict((dim, sources.coords[dim]) for dim in sources.dims),
-                                       geobox, measurements, data_func)
+                                       geobox, measurements, data_func, use_threads)
 
     @staticmethod
     def measurement_data(sources, geobox, measurement, fuse_func=None, dask_chunks=None):
         """
         Retrieve a single measurement variable as a :class:`xarray.DataArray`.
 
+        .. note:
+
+             This method appears to only be used by the deprecated `get_data()/get_descriptor()`
+              :class:`~datacube.api.API`, so is a prime candidate for future removal.
+
+        .. seealso:: :meth:`load_data`
+
+
         :param xarray.DataArray sources: DataArray holding a list of :class:`datacube.model.Dataset` objects
         :param GeoBox geobox: A GeoBox defining the output spatial projection and resolution
         :param measurement: measurement definition with keys: {'name', 'dtype', 'nodata', 'units'}
         :param fuse_func: function to merge successive arrays as an output
         :param dict dask_chunks: If the data should be loaded as needed using :class:`dask.array.Array`,
             specify the chunk size in each output direction.
-
             See the documentation on using `xarray with dask <http://xarray.pydata.org/en/stable/dask.html>`_
             for more information.
         :rtype: :class:`xarray.DataArray`
-
-        .. seealso:: :meth:`load_data`
         """
         dataset = Datacube.load_data(sources, geobox, [measurement], fuse_func=fuse_func, dask_chunks=dask_chunks)
         dataarray = dataset[measurement['name']]
@@ -558,6 +578,43 @@ class Datacube(object):
         self.close()
 
 
+def output_geobox(like=None, output_crs=None, resolution=None, align=None,
+                  grid_spec=None, datasets=None, **query):
+    """ Configure output geobox from user provided output specs. """
+
+    if like is not None:
+        assert output_crs is None, "'like' and 'output_crs' are not supported together"
+        assert resolution is None, "'like' and 'resolution' are not supported together"
+        assert align is None, "'like' and 'align' are not supported together"
+        return like.geobox
+
+    if output_crs is not None:
+        # user provided specifications
+        if resolution is None:
+            raise ValueError("Must specify 'resolution' when specifying 'output_crs'")
+        crs = geometry.CRS(output_crs)
+    else:
+        # specification from grid_spec
+        if grid_spec is None or grid_spec.crs is None:
+            raise ValueError("Product has no default CRS. Must specify 'output_crs' and 'resolution'")
+        crs = grid_spec.crs
+        if resolution is None:
+            if grid_spec.resolution is None:
+                raise ValueError("Product has no default resolution. Must specify 'resolution'")
+            resolution = grid_spec.resolution
+        align = align or grid_spec.alignment
+
+    return geometry.GeoBox.from_geopolygon(query_geopolygon(**query) or get_bounds(datasets, crs),
+                                           resolution, crs, align)
+
+
+def select_datasets_inside_polygon(datasets, polygon):
+    # Check against the bounding box of the original scene, can throw away some portions
+    for dataset in datasets:
+        if polygon is None or intersects(polygon.to_crs(dataset.crs), dataset.extent):
+            yield dataset
+
+
 def fuse_lazy(datasets, geobox, measurement, skip_broken_datasets=False, fuse_func=None, prepend_dims=0):
     prepend_shape = (1,) * prepend_dims
     data = numpy.full(geobox.shape, measurement['nodata'], dtype=measurement['dtype'])
@@ -567,8 +624,10 @@ def fuse_lazy(datasets, geobox, measurement, skip_broken_datasets=False, fuse_fu
     return data.reshape(prepend_shape + geobox.shape)
 
 
-def _fuse_measurement(dest, datasets, geobox, measurement, skip_broken_datasets=False, fuse_func=None):
-    reproject_and_fuse([RasterDatasetSource(dataset, measurement['name']) for dataset in datasets],
+def _fuse_measurement(dest, datasets, geobox, measurement,
+                      skip_broken_datasets=False,
+                      fuse_func=None):
+    reproject_and_fuse([new_datasource(dataset, measurement['name']) for dataset in datasets],
                        dest,
                        geobox.affine,
                        geobox.crs,
@@ -600,7 +659,7 @@ def set_resampling_method(measurements, resampling=None):
     return measurements
 
 
-def datatset_type_to_row(dt):
+def dataset_type_to_row(dt):
     row = {
         'id': dt.id,
         'name': dt.name,
@@ -609,7 +668,7 @@ def datatset_type_to_row(dt):
     row.update(dt.fields)
     if dt.grid_spec is not None:
         row.update({
-            'crs': dt.grid_spec.crs,
+            'crs': str(dt.grid_spec.crs),
             'spatial_dimensions': dt.grid_spec.dimensions,
             'tile_size': dt.grid_spec.tile_size,
             'resolution': dt.grid_spec.resolution,
@@ -618,10 +677,10 @@ def datatset_type_to_row(dt):
 
 
 def _chunk_geobox(geobox, chunk_size):
-    num_grid_chunks = [int(ceil(s/float(c))) for s, c in zip(geobox.shape, chunk_size)]
+    num_grid_chunks = [int(ceil(s / float(c))) for s, c in zip(geobox.shape, chunk_size)]
     geobox_subsets = {}
     for grid_index in numpy.ndindex(*num_grid_chunks):
-        slices = [slice(min(d*c, stop), min((d+1)*c, stop))
+        slices = [slice(min(d * c, stop), min((d + 1) * c, stop))
                   for d, c, stop in zip(grid_index, chunk_size, geobox.shape)]
         geobox_subsets[grid_index] = geobox[slices]
     return geobox_subsets
@@ -631,7 +690,7 @@ def _calculate_chunk_sizes(sources, geobox, dask_chunks):
     valid_keys = sources.dims + geobox.dimensions
     bad_keys = set(dask_chunks) - set(valid_keys)
     if bad_keys:
-        raise KeyError('Unknown dask_chunk dimension {}. Valid dimensions are: {}', bad_keys, valid_keys)
+        raise KeyError('Unknown dask_chunk dimension {}. Valid dimensions are: {}'.format(bad_keys, valid_keys))
 
     # If chunk size is not specified, the entire dimension length is used, as in xarray
     chunks = {dim: size for dim, size in zip(sources.dims, sources.shape)}
@@ -645,8 +704,11 @@ def _calculate_chunk_sizes(sources, geobox, dask_chunks):
     return irr_chunks, grid_chunks
 
 
-def _make_dask_array(sources, geobox, measurement, skip_broken_datasets=False, fuse_func=None, dask_chunks=None):
-    #: pylint: disable=too-many-locals
+# pylint: disable=too-many-locals
+def _make_dask_array(sources, geobox, measurement,
+                     skip_broken_datasets=False,
+                     fuse_func=None,
+                     dask_chunks=None):
     dsk_name = 'datacube_' + measurement['name']
 
     irr_chunks, grid_chunks = _calculate_chunk_sizes(sources, geobox, dask_chunks)
