@@ -81,17 +81,20 @@ class Tile(object):
         geobox = self.geobox[chunk[len(self.sources.shape):]]
         return Tile(sources, geobox)
 
-    def split(self, dim):
+    # TODO(csiro) Split on time range
+    def split(self, dim, step=1):
         """
-        Splits along a non-spatial dimension into Tile objects with a length of 1 in the `dim` dimension.
+        Splits along a non-spatial dimension into Tile objects with a length of 1 or more in the `dim` dimension.
 
         :param dim: Name of the non-spatial dimension to split
+        :param step: step size to split
         :return: tuple(key, Tile)
         """
         axis = self.dims.index(dim)
         indexer = [slice(None)] * len(self.dims)
-        for i in range(self.sources[dim].size):
-            indexer[axis] = slice(i, i + 1)
+        size = self.sources[dim].size
+        for i in range(0, size, step):
+            indexer[axis] = slice(i, min(size, i + step))
             yield self.sources[dim].values[i], self[tuple(indexer)]
 
     def split_by_time(self, freq='A', time_dim='time', **kwargs):
@@ -142,7 +145,7 @@ class GridWorkflow(object):
 
         Either grid_spec or product must be supplied.
 
-        :param Index index: The database index to use.
+        :param datacube.index.Index index: The database index to use.
         :param GridSpec grid_spec: The grid projection and resolution
         :param str product: The name of an existing product, if no grid_spec is supplied.
         """
@@ -360,19 +363,18 @@ class GridWorkflow(object):
 
             Defaults to ``'nearest'``.
 
+        :param bool skip_broken_datasets: If True, ignore broken datasets and continue processing with the data
+             that can be loaded. If False, an exception will be raised on a broken dataset. Defaults to False.
+
         :rtype: :py:class:`xarray.Dataset`
 
         .. seealso::
             :meth:`list_tiles` :meth:`list_cells`
         """
-        observations = []
-        for dataset in tile.sources.values:
-            observations += dataset
-
         measurements = tile.product.lookup_measurements(measurements)
         measurements = set_resampling_method(measurements, resampling)
 
-        dataset = Datacube.load_data(tile.sources, tile.geobox, measurements.values(), dask_chunks=dask_chunks,
+        dataset = Datacube.load_data(tile.sources, tile.geobox, list(measurements.values()), dask_chunks=dask_chunks,
                                      fuse_func=fuse_func, skip_broken_datasets=skip_broken_datasets)
 
         return dataset
