@@ -1,4 +1,3 @@
-
 from __future__ import absolute_import
 
 
@@ -10,5 +9,5 @@ class MissingRecordError(Exception):
     pass
 
 
-class UnknownFieldError(Exception):
+class IndexSetupError(Exception):
     pass
