@@ -5,10 +5,7 @@ import logging
 import click
 import cachetools
 import itertools
-try:
-    import cPickle as pickle
-except ImportError:
-    import pickle
+import sys
 from copy import deepcopy
 from pathlib import Path
 from pandas import to_datetime
@@ -16,16 +13,17 @@ from datetime import datetime
 
 import datacube
 from datacube.api.core import Datacube
+from datacube.index.index import Index
 from datacube.model import DatasetType, Range, GeoPolygon
 from datacube.model.utils import make_dataset, xr_apply, datasets_to_doc
-from datacube.storage.storage import write_dataset_to_netcdf
 from datacube.ui import click as ui
-from datacube.utils import read_documents, changes
+from datacube.utils import read_documents
 from datacube.ui.task_app import check_existing_files, load_tasks as load_tasks_, save_tasks as save_tasks_
+from datacube.drivers import storage_writer_by_name
 
 from datacube.ui.click import cli
 
-_LOG = logging.getLogger('agdc-ingest')
+_LOG = logging.getLogger('datacube-ingest')
 
 FUSER_KEY = 'fuse_data'
 
@@ -41,7 +39,7 @@ def find_diff(input_type, output_type, index, **query):
     return tasks
 
 
-def morph_dataset_type(source_type, config, index):
+def morph_dataset_type(source_type, config, index, storage_format):
     output_metadata_type = source_type.metadata_type
     if 'metadata_type' in config:
         output_metadata_type = index.metadata_types.get_by_name(config['metadata_type'])
@@ -50,11 +48,11 @@ def morph_dataset_type(source_type, config, index):
     output_type.definition['name'] = config['output_type']
     output_type.definition['managed'] = True
     output_type.definition['description'] = config['description']
-    output_type.metadata_doc['format'] = {'name': 'NetCDF'}
-
     output_type.definition['storage'] = {k: v for (k, v) in config['storage'].items()
                                          if k in ('crs', 'tile_size', 'resolution', 'origin')}
 
+    output_type.metadata_doc['format'] = {'name': storage_format}
+
     if 'metadata_type' in config:
         output_type.definition['metadata_type'] = config['metadata_type']
 
@@ -90,8 +88,7 @@ def get_app_metadata(config, config_file):
         'lineage': {
             'algorithm': {
                 'name': 'datacube-ingest',
-                'version': config.get('version', 'unknown'),
-                'repo_url': 'https://github.com/GeoscienceAustralia/datacube-ingester.git',
+                'repo_url': 'https://github.com/opendatacube/datacube-core.git',
                 'parameters': {'configuration_file': config_file}
             },
         }
@@ -123,7 +120,7 @@ def get_namemap(config):
     return {spec['src_varname']: spec['name'] for spec in config['measurements']}
 
 
-def ensure_output_type(index, config, allow_product_changes=False):
+def ensure_output_type(index, config, storage_format, allow_product_changes=False):
     # type: (Index, dict, bool) -> (DatasetType, DatasetType)
     """
     Create the output product for the given ingest config if it doesn't already exist.
@@ -136,7 +133,7 @@ def ensure_output_type(index, config, allow_product_changes=False):
         click.echo("Source DatasetType %s does not exist" % config['source_type'])
         click.get_current_context().exit(1)
 
-    output_type = morph_dataset_type(source_type, config, index)
+    output_type = morph_dataset_type(source_type, config, index, storage_format)
     _LOG.info('Created DatasetType %s', output_type.name)
 
     existing = index.products.get_by_name(output_type.name)
@@ -158,10 +155,10 @@ def get_full_lineage(index, id_):
     return index.datasets.get(id_, include_sources=True)
 
 
-def load_config_from_file(index, config):
-    config_name = Path(config).name
-    _, config = next(read_documents(Path(config)))
-    config['filename'] = config_name
+def load_config_from_file(path):
+    config_file = Path(path)
+    _, config = next(read_documents(config_file))
+    config['filename'] = str(config_file.absolute())
 
     return config
 
@@ -204,7 +201,14 @@ def create_task_list(index, output_type, year, source_type, config):
 
 
 def ingest_work(config, source_type, output_type, tile, tile_index):
+    # pylint: disable=too-many-locals
     _LOG.info('Starting task %s', tile_index)
+    driver = storage_writer_by_name(config['storage']['driver'])
+
+    if driver is None:
+        _LOG.error('Failed to load storage driver %s', config['storage']['driver'])
+        raise ValueError('Something went wrong: no longer can find driver pointed by storage.driver option')
+
     namemap = get_namemap(config)
     measurements = get_measurements(source_type, config)
     variable_params = get_variable_params(config)
@@ -213,22 +217,41 @@ def ingest_work(config, source_type, output_type, tile, tile_index):
     with datacube.set_options(reproject_threads=1):
         fuse_func = {'copy': None}[config.get(FUSER_KEY, 'copy')]
         data = Datacube.load_data(tile.sources, tile.geobox, measurements, fuse_func=fuse_func)
+
     nudata = data.rename(namemap)
     file_path = get_filename(config, tile_index, tile.sources)
 
+    def mk_uri(file_path):
+        if driver.uri_scheme == "file":
+            return file_path.absolute().as_uri()
+        return '{}://{}'.format(driver.uri_scheme, file_path)
+
     def _make_dataset(labels, sources):
         return make_dataset(product=output_type,
                             sources=sources,
                             extent=tile.geobox.extent,
                             center_time=labels['time'],
-                            uri=file_path.absolute().as_uri(),
+                            uri=mk_uri(file_path),
                             app_info=get_app_metadata(config, config['filename']),
                             valid_data=GeoPolygon.from_sources_extents(sources, tile.geobox))
 
     datasets = xr_apply(tile.sources, _make_dataset, dtype='O')  # Store in Dataarray to associate Time -> Dataset
     nudata['dataset'] = datasets_to_doc(datasets)
 
-    write_dataset_to_netcdf(nudata, file_path, global_attributes, variable_params)
+    variable_params['dataset'] = {
+        'chunksizes': (1,),
+        'zlib': True,
+        'complevel': 9,
+    }
+
+    storage_metadata = driver.write_dataset_to_storage(nudata, file_path,
+                                                       global_attributes=global_attributes,
+                                                       variable_params=variable_params,
+                                                       storage_config=config['storage'])
+
+    if (storage_metadata is not None) and len(storage_metadata) > 0:
+        datasets.attrs['storage_metadata'] = storage_metadata
+
     _LOG.info('Finished task %s', tile_index)
 
     return datasets
@@ -237,13 +260,19 @@ def ingest_work(config, source_type, output_type, tile, tile_index):
 def _index_datasets(index, results):
     n = 0
     for datasets in results:
+        extra_args = {}
+        # datasets is an xarray.DataArray
+        if 'storage_metadata' in datasets.attrs:
+            extra_args['storage_metadata'] = datasets.attrs['storage_metadata']
+
         for dataset in datasets.values:
-            index.datasets.add(dataset, sources_policy='skip')
+            index.datasets.add(dataset, sources_policy='skip', **extra_args)
             n += 1
     return n
 
 
 def process_tasks(index, config, source_type, output_type, tasks, queue_size, executor):
+    # pylint: disable=too-many-locals
     def submit_task(task):
         _LOG.info('Submitting task: %s', task['tile_index'])
         return executor.submit(ingest_work,
@@ -257,7 +286,7 @@ def process_tasks(index, config, source_type, output_type, tasks, queue_size, ex
 
     tasks = iter(tasks)
     while True:
-        pending += [submit_task(task) for task in itertools.islice(tasks, max(0, queue_size-len(pending)))]
+        pending += [submit_task(task) for task in itertools.islice(tasks, max(0, queue_size - len(pending)))]
         if not pending:
             break
 
@@ -279,8 +308,9 @@ def process_tasks(index, config, source_type, output_type, tasks, queue_size, ex
             # TODO: ideally we wouldn't block here indefinitely
             # maybe limit gather to 50-100 results and put the rest into a index backlog
             # this will also keep the queue full
-            n_successful += _index_datasets(index, executor.results(completed))
-        except Exception:  # pylint: disable=broad-except
+            results = executor.results(completed)
+            n_successful += _index_datasets(index, results)
+        except Exception as e:  # pylint: disable=broad-except
             _LOG.exception('Gather failed')
             pending += completed
 
@@ -300,6 +330,15 @@ def _validate_year(ctx, param, value):
                                  'or as an inclusive range (eg 1996-2001)')
 
 
+def get_driver_from_config(config):
+    driver_name = config['storage']['driver']
+    driver = storage_writer_by_name(driver_name)
+    if driver is None:
+        click.echo('Failed to load requested storage driver: ' + driver_name)
+        sys.exit(2)
+    return driver
+
+
 @cli.command('ingest', help="Ingest datasets")
 @click.option('--config-file', '-c',
               type=click.Path(exists=True, readable=True, writable=False, dir_okay=False),
@@ -314,7 +353,7 @@ def _validate_year(ctx, param, value):
 @click.option('--allow-product-changes', is_flag=True, default=False,
               help='Allow the output product definition to be updated if it differs.')
 @ui.executor_cli_options
-@ui.pass_index(app_name='agdc-ingest')
+@ui.pass_index(app_name='datacube-ingest')
 def ingest_cmd(index,
                config_file,
                year,
@@ -324,26 +363,30 @@ def ingest_cmd(index,
                dry_run,
                executor,
                allow_product_changes):
+    # pylint: disable=too-many-locals
+
     if config_file:
-        config = load_config_from_file(index, config_file)
-        source_type, output_type = ensure_output_type(index, config, allow_product_changes=allow_product_changes)
+        config = load_config_from_file(config_file)
+        driver = get_driver_from_config(config)
+        source_type, output_type = ensure_output_type(index, config, driver.format,
+                                                      allow_product_changes=allow_product_changes)
 
         tasks = create_task_list(index, output_type, year, source_type, config)
     elif load_tasks:
         config, tasks = load_tasks_(load_tasks)
-        source_type, output_type = ensure_output_type(index, config, allow_product_changes=allow_product_changes)
+        driver = get_driver_from_config(config)
+        source_type, output_type = ensure_output_type(index, config, driver.format,
+                                                      allow_product_changes=allow_product_changes)
     else:
         click.echo('Must specify exactly one of --config-file, --load-tasks')
-        return 1
+        sys.exit(-1)
 
     if dry_run:
         check_existing_files(get_filename(config, task['tile_index'], task['tile'].sources) for task in tasks)
-        return 0
-
-    if save_tasks:
+    elif save_tasks:
         save_tasks_(config, tasks, save_tasks)
-        return 0
+    else:
+        successful, failed = process_tasks(index, config, source_type, output_type, tasks, queue_size, executor)
+        click.echo('%d successful, %d failed' % (successful, failed))
 
-    successful, failed = process_tasks(index, config, source_type, output_type, tasks, queue_size, executor)
-    click.echo('%d successful, %d failed' % (successful, failed))
-    return 0
+        sys.exit(failed)
