@@ -15,20 +15,17 @@ import math
 from contextlib import contextmanager
 from pathlib import Path
 
-from datacube.compat import urlparse, urljoin, url_parse_module
+import urllib.parse
+from urllib.parse import urlparse, urljoin
 from datacube.config import OPTIONS
+from datacube.drivers.datasource import DataSource
 from datacube.model import Dataset
 from datacube.storage import netcdf_writer
 from datacube.utils import clamp, datetime_to_seconds_since_1970, DatacubeException, ignore_exceptions_if
 from datacube.utils import geometry
-from datacube.utils import is_url, uri_to_local_path
+from datacube.utils import is_url, uri_to_local_path, get_part_from_uri
 
-try:
-    from yaml import CSafeDumper as SafeDumper
-except ImportError:
-    from yaml import SafeDumper
 import numpy
-
 from affine import Affine
 from datacube.compat import integer_types
 import rasterio
@@ -54,6 +51,11 @@ GDAL_NETCDF_DIM = ('NETCDF_DIM_'
                    if str(rasterio.__gdal_version__) >= '1.10.0' else
                    'NETCDF_DIMENSION_')
 
+# rasterio 1.0aX series had warp bug until 1.0a9 So we use workaround code for
+# these versions only, because of silly numbering scheme we can't use simple
+# comparison operators, since '1.0a10' is less than '1.0a9'
+RASTERIO_WARP_BUG = (str(rasterio.__version__) in set('1.0a%d' % i for i in range(9)))
+
 
 def _rasterio_resampling_method(resampling):
     return RESAMPLING_METHODS[resampling.lower()]
@@ -78,32 +80,33 @@ else:
 
 
 def _calc_offsets_impl(off, scale, src_size, dst_size):
-    assert scale >= 1-1e-5
+    assert scale >= 1 - 1e-5
 
     if off >= 0:
         write_off = 0
     else:
-        write_off = math.ceil((-off-0.5)/scale)
-    read_off = round((write_off+0.5)*scale-0.5+off) - round(0.5*(scale-1.0))  # assuming read_size/write_size ~= scale
+        write_off = math.ceil((-off - 0.5) / scale)
+    read_off = round((write_off + 0.5) * scale - 0.5 + off) - round(
+        0.5 * (scale - 1.0))  # assuming read_size/write_size ~= scale
     if read_off >= src_size:
         return 0, 0, 0, 0
 
     write_end = dst_size
-    write_size = write_end-write_off
-    read_end = read_off+round(write_size*scale)
+    write_size = write_end - write_off
+    read_end = read_off + round(write_size * scale)
     if read_end > src_size:
         # +0.5 below is a fudge that will return last row in more situations, but will change the scale more
-        write_end = math.floor((src_size-off+0.5)/scale)
-        write_size = write_end-write_off
-        read_end = clamp(read_off+round(write_size*scale), read_off, src_size)
-    read_size = read_end-read_off
+        write_end = math.floor((src_size - off + 0.5) / scale)
+        write_size = write_end - write_off
+        read_end = clamp(read_off + round(write_size * scale), read_off, src_size)
+    read_size = read_end - read_off
 
     return int(read_off), int(write_off), int(read_size), int(write_size)
 
 
 def _calc_offsets2(off, scale, src_size, dst_size):
     if scale < 0:
-        r_off, write_off, read_size, write_size = _calc_offsets_impl(off + dst_size*scale, -scale, src_size, dst_size)
+        r_off, write_off, read_size, write_size = _calc_offsets_impl(off + dst_size * scale, -scale, src_size, dst_size)
         return r_off, dst_size - write_size - write_off, read_size, write_size
     else:
         return _calc_offsets_impl(off, scale, src_size, dst_size)
@@ -113,11 +116,11 @@ def _read_decimated(array_transform, src, dest_shape):
     dy_dx = (array_transform.f, array_transform.c)
     sy_sx = (array_transform.e, array_transform.a)
     read, write, read_shape, write_shape = zip(*map(_calc_offsets2, dy_dx, sy_sx, src.shape, dest_shape))
-    if all(write_shape):
+    if write_shape[0] > 0 and write_shape[1] > 0 and all(write_shape):
         window = ((read[0], read[0] + read_shape[0]), (read[1], read[1] + read_shape[1]))
         tmp = src.read(window=window, out_shape=write_shape)
-        scale = (read_shape[0]/write_shape[0] if sy_sx[0] > 0 else -read_shape[0]/write_shape[0],
-                 read_shape[1]/write_shape[1] if sy_sx[1] > 0 else -read_shape[1]/write_shape[1])
+        scale = (read_shape[0] / write_shape[0] if sy_sx[0] > 0 else -read_shape[0] / write_shape[0],
+                 read_shape[1] / write_shape[1] if sy_sx[1] > 0 else -read_shape[1] / write_shape[1])
         offset = (read[0] + (0 if sy_sx[0] > 0 else read_shape[0]),
                   read[1] + (0 if sy_sx[1] > 0 else read_shape[1]))
         transform = Affine(scale[1], 0, offset[1], 0, scale[0], offset[0])
@@ -169,12 +172,12 @@ def read_from_source(source, dest, dst_transform, dst_nodata, dst_projection, re
                           NUM_THREADS=OPTIONS['reproject_threads'])
 
 
-def reproject_and_fuse(sources, destination, dst_transform, dst_projection, dst_nodata,
+def reproject_and_fuse(datasources, destination, dst_transform, dst_projection, dst_nodata,
                        resampling='nearest', fuse_func=None, skip_broken_datasets=False):
     """
     Reproject and fuse `sources` into a 2D numpy array `destination`.
 
-    :param List[BaseRasterDataSource] sources: Data sources to open and read from
+    :param List[DataSource] datasources: Data sources to open and read from
     :param numpy.ndarray destination: ndarray of appropriate size to read data into
     :type resampling: str
     :type fuse_func: callable or None
@@ -194,16 +197,16 @@ def reproject_and_fuse(sources, destination, dst_transform, dst_projection, dst_
     fuse_func = fuse_func or copyto_fuser
 
     destination.fill(dst_nodata)
-    if len(sources) == 0:
+    if len(datasources) == 0:
         return destination
-    elif len(sources) == 1:
+    elif len(datasources) == 1:
         with ignore_exceptions_if(skip_broken_datasets):
-            read_from_source(sources[0], destination, dst_transform, dst_nodata, dst_projection, resampling)
+            read_from_source(datasources[0], destination, dst_transform, dst_nodata, dst_projection, resampling)
         return destination
     else:
-        # Muitiple sources, we need to fuse them together into a single array
+        # Multiple sources, we need to fuse them together into a single array
         buffer_ = numpy.empty(destination.shape, dtype=destination.dtype)
-        for source in sources:
+        for source in datasources:
             with ignore_exceptions_if(skip_broken_datasets):
                 read_from_source(source, buffer_, dst_transform, dst_nodata, dst_projection, resampling)
                 fuse_func(destination, buffer_)
@@ -217,6 +220,7 @@ class BandDataSource(object):
 
     :type source: rasterio.Band
     """
+
     def __init__(self, source, nodata=None):
         self.source = source
         if nodata is None:
@@ -256,80 +260,15 @@ class BandDataSource(object):
                                        **kwargs)
 
 
-# class NetCDFDataSource(object):
-#     def __init__(self, dataset, variable, slab=None, nodata=None):
-#         self.dataset = dataset
-#         self.variable = self.dataset[variable]
-#         self.slab = slab or {}
-#         if nodata is None:
-#             nodata = self.variable.getncattr('_FillValue')
-#         self.nodata = nodata
-#
-#     @property
-#     def crs(self):
-#         crs_var_name = self.variable.grid_mapping
-#         crs_var = self.dataset[crs_var_name]
-#         return geometry.CRS(crs_var.crs_wkt)
-#
-#     @property
-#     def transform(self):
-#         dims = self.crs.dimensions
-#         xres, xoff = data_resolution_and_offset(self.dataset[dims[1]])
-#         yres, yoff = data_resolution_and_offset(self.dataset[dims[0]])
-#         return Affine.translation(xoff, yoff) * Affine.scale(xres, yres)
-#
-#     @property
-#     def dtype(self):
-#         return self.variable.dtype
-#
-#     @property
-#     def shape(self):
-#         return self.variable.shape
-#
-#     def read(self, window=None, out_shape=None):
-#         data = self.variable
-#         if window is None:
-#             window = ((0, data.shape[0]), (0, data.shape[1]))
-#         data_shape = (window[0][1]-window[0][0]), (window[1][1]-window[1][0])
-#         if out_shape is None:
-#             out_shape = data_shape
-#         xidx = window[0][0] + ((
-# numpy.arange(out_shape[1])+0.5)*(data_shape[1]/out_shape[1])-0.5).round().astype('int')
-#         yidx = window[1][0] + ((
-# numpy.arange(out_shape[0])+0.5)*(data_shape[0]/out_shape[0])-0.5).round().astype('int')
-#         slab = {self.crs.dimensions[1]: xidx, self.crs.dimensions[0]: yidx}
-#         slab.update(self.slab)
-#         return data[tuple(slab[d] for d in self.variable.dimensions)]
-#
-#     def reproject(self, dest, dst_transform, dst_crs, dst_nodata, resampling, **kwargs):
-#         dst_poly = geometry.polygon_from_transform(dest.shape[1], dest.shape[0],
-#                                                    dst_transform, dst_crs).to_crs(self.crs)
-#         src_poly = geometry.polygon_from_transform(self.shape[1], self.shape[0],
-#                                                    self.transform, self.crs)
-#         bounds = dst_poly.intersection(src_poly)
-#         geobox = geometry.GeoBox.from_geopolygon(bounds, (self.transform.e, self.transform.a), crs=self.crs)
-#         tmp, _, tmp_transform = _read_decimated(~self.transform * geobox.affine, self, geobox.shape)
-#
-#         return rasterio.warp.reproject(tmp,
-#                                        dest,
-#                                        src_transform=self.transform * tmp_transform,
-#                                        src_crs=str(geobox.crs),
-#                                        src_nodata=self.nodata,
-#                                        dst_transform=dst_transform,
-#                                        dst_crs=str(dst_crs),
-#                                        dst_nodata=dst_nodata,
-#                                        resampling=resampling,
-#                                        **kwargs)
-
-
 class OverrideBandDataSource(object):
-    """Wrapper for a rasterio.Band object that overrides nodata, crs and transform
+    """Wrapper for a rasterio.Band object that overrides nodata, CRS and transform
 
     This is useful for files with malformed or missing properties.
 
 
     :type source: rasterio.Band
     """
+
     def __init__(self, source, nodata, crs, transform):
         self.source = source
         self.nodata = nodata
@@ -363,11 +302,12 @@ class OverrideBandDataSource(object):
                                        **kwargs)
 
 
-class RasterioDataSource(object):
+class RasterioDataSource(DataSource):
     """
     Abstract class used by fuse_sources and :func:`read_from_source`
 
     """
+
     def __init__(self, filename, nodata):
         self.filename = filename
         self.nodata = nodata
@@ -400,12 +340,12 @@ class RasterioDataSource(object):
                     override = True
                     crs = self.get_crs()
 
-                # The 1.0 onwards release of rasterio has a bug that means it
+                # The [1.0a1-1.0a8] releases of rasterio had a bug that means it
                 # cannot read multiband data into a numpy array during reprojection
                 # We override it here to force the reading and reprojection into separate steps
-                # TODO: Remove when rasterio bug fixed
+                # TODO: Remove when we no longer care about those versions of rasterio
                 bandnumber = self.get_bandnumber(src)
-                if bandnumber > 1 and str(rasterio.__version__) >= '1.0':
+                if bandnumber > 1 and RASTERIO_WARP_BUG:
                     override = True
 
                 band = rasterio.band(src, bandnumber)
@@ -443,7 +383,7 @@ class RasterFileDataSource(RasterioDataSource):
         return self.crs
 
 
-class RasterDatasetSource(RasterioDataSource):
+class RasterDatasetDataSource(RasterioDataSource):
     """Data source for reading from a Data Cube Dataset"""
 
     def __init__(self, dataset, measurement_id):
@@ -455,12 +395,14 @@ class RasterDatasetSource(RasterioDataSource):
         """
         self._dataset = dataset
         self._measurement = dataset.measurements[measurement_id]
+        self._netcdf = ('netcdf' in dataset.format.lower())
         url = _resolve_url(_choose_location(dataset), self._measurement['path'])
+        self._part = get_part_from_uri(url)
         filename = _url2rasterio(url, dataset.format, self._measurement.get('layer'))
         nodata = dataset.type.measurements[measurement_id].get('nodata')
-        super(RasterDatasetSource, self).__init__(filename, nodata=nodata)
+        super(RasterDatasetDataSource, self).__init__(filename, nodata=nodata)
 
-    def get_bandnumber(self, src):
+    def get_bandnumber(self, src=None):
 
         # If `band` property is set to an integer it overrides any other logic
         band = self._measurement.get('band')
@@ -470,10 +412,28 @@ class RasterDatasetSource(RasterioDataSource):
             else:
                 _LOG.warning('Expected "band" property to be of integer type')
 
-        if 'netcdf' not in self._dataset.format.lower():
+        if not self._netcdf:
             layer_id = self._measurement.get('layer', 1)
             return layer_id if isinstance(layer_id, integer_types) else 1
 
+        # Netcdf only below
+        if self._part is not None:
+            return self._part + 1  # Convert to rasterio 1-based indexing
+
+        if src is None:
+            # File wasnt' open, could be unstacked file in a new format, or
+            # stacked/unstacked in old. We assume caller knows what to do
+            # (maybe based on some side-channel information), so just report
+            # undefined.
+            return None
+
+        if src.count == 1:  # Single-slice netcdf file
+            return 1
+
+        _LOG.debug("Encountered stacked netcdf file without recorded index\n - %s", src.name)
+
+        # Below is backwards compatibility code
+
         tag_name = GDAL_NETCDF_DIM + 'time'
         if tag_name not in src.tags(1):  # TODO: support time-less datasets properly
             return 1
@@ -491,7 +451,7 @@ class RasterDatasetSource(RasterioDataSource):
         return idx
 
     def get_transform(self, shape):
-        return self._dataset.transform * Affine.scale(1/shape[1], 1/shape[0])
+        return self._dataset.transform * Affine.scale(1 / shape[1], 1 / shape[0])
 
     def get_crs(self):
         return self._dataset.crs
@@ -502,9 +462,10 @@ def register_scheme(*schemes):
     Register additional uri schemes as supporting relative offsets (etc), so that band/measurement paths can be
     calculated relative to the base uri.
     """
-    url_parse_module.uses_netloc.extend(schemes)
-    url_parse_module.uses_relative.extend(schemes)
-    url_parse_module.uses_params.extend(schemes)
+    urllib.parse.uses_netloc.extend(schemes)
+    urllib.parse.uses_relative.extend(schemes)
+    urllib.parse.uses_params.extend(schemes)
+
 
 # Not recognised by python by default. Doctests below will fail without it.
 register_scheme('s3')
@@ -582,6 +543,16 @@ def _choose_location(dataset):
     return uris[0]
 
 
+def measurement_paths(dataset):
+    '''Returns a dictionary mapping from band name to url pointing to band storage
+    resource.
+    :return: {str: str} Band Name => URL
+    '''
+    base = _choose_location(dataset)
+    return dict((k, _resolve_url(base, m.get('path', '')))
+                for k, m in dataset.measurements.items())
+
+
 def create_netcdf_storage_unit(filename,
                                crs, coordinates, variables, variable_params, global_attributes=None,
                                netcdfparams=None):
